@@ -80,16 +80,7 @@ When called interactively, a Tramp connection has to be selected."
    ;; Return nil when there is no Tramp connection.
    (list
     (let ((connections
-	   (mapcar
-	    (lambda (x)
-	      (tramp-make-tramp-file-name
-	       (tramp-file-name-method x)
-	       (tramp-file-name-user x)
-	       (tramp-file-name-domain x)
-	       (tramp-file-name-host x)
-	       (tramp-file-name-port x)
-	       (tramp-file-name-localname x)))
-	    (tramp-list-connections)))
+	   (mapcar 'tramp-make-tramp-file-name (tramp-list-connections)))
 	  name)
 
       (when connections
@@ -113,13 +104,13 @@ When called interactively, a Tramp connection has to be selected."
     (when keep-password (setq tramp-current-connection nil))
 
     ;; Flush file cache.
-    (tramp-flush-directory-property vec "")
+    (tramp-flush-directory-properties vec "")
 
     ;; Flush connection cache.
     (when (processp (tramp-get-connection-process vec))
-      (tramp-flush-connection-property (tramp-get-connection-process vec))
+      (tramp-flush-connection-properties (tramp-get-connection-process vec))
       (delete-process (tramp-get-connection-process vec)))
-    (tramp-flush-connection-property vec)
+    (tramp-flush-connection-properties vec)
 
     ;; Remove buffers.
     (dolist
@@ -152,6 +143,10 @@ This includes password cache, file cache, connection cache, buffers."
   ;; Flush file and connection cache.
   (clrhash tramp-cache-data)
 
+  ;; Cleanup local copies of archives.
+  (when (bound-and-true-p tramp-archive-enabled)
+    (tramp-archive-cleanup-hash))
+
   ;; Remove buffers.
   (dolist (name (tramp-list-tramp-buffers))
     (when (bufferp (get-buffer name)) (kill-buffer name))))
@@ -186,7 +181,9 @@ This includes password cache, file cache, connection cache, buffers."
   "Submit a bug report to the Tramp developers."
   (interactive)
   (catch 'dont-send
-    (let ((reporter-prompt-for-summary-p t))
+    (let ((reporter-prompt-for-summary-p t)
+	  ;; In rare cases, it could contain the password.  So we make it nil.
+	  tramp-password-save-function)
       (reporter-submit-bug-report
        tramp-bug-report-address		; to-address
        (format "tramp (%s)" tramp-version) ; package name and version
